
package smsqesprites;

/**
 * Makes colour reductions to SMSQE colour modes.
 * @author and copyright (c) wolfgang Lenerz 20132
 */
public class ColourReduction 
{ 
    /**
    * Converts the source image to 2-bit colour (SMSQ/E mode 4).
    * 
    * @param src the source image to convert.
    * 
    * @return a new image with a 2-bit colour depth
    */
    public static java.awt.image.BufferedImage reduceToMode4 (java.awt.image.BufferedImage src)
    {
        // make a custom palette
        java.awt.image.IndexColorModel cm = new java.awt.image.IndexColorModel(
                2, // 4 bits can store up to 4 colors
                4, // I use all of them          
                new byte[]{ -1,   0,   -1,     0},
                new byte[]{  0,  -1,   -1,     0},
                new byte[]{  0,  0,    -1,     0});
                //          RED  GREEN WHITE BLACK    

        // draw source image onto intermediate, with custom palette
        java.awt.image.BufferedImage interm = new java.awt.image.BufferedImage(
                src.getWidth(), src.getHeight(), // match source
                java.awt.image.BufferedImage.TYPE_BYTE_INDEXED, 
                cm); 
        java.awt.Graphics2D g2 = interm.createGraphics();
        g2.drawImage(src, 0, 0, null);
        g2.dispose();
        // draw intermediate image onto destination
  /*      java.awt.image.BufferedImage dest = new java.awt.image.BufferedImage(
                src.getWidth(), src.getHeight(), // match source
                java.awt.image.BufferedImage.TYPE_INT_ARGB); 
        g2 = dest.createGraphics();
        g2.drawImage(interm, 0, 0, null);
        g2.dispose();
   //     copyAlpha(src,dest);
   */
        return interm;
    }
    
    
   /**
    * Converts the source image to 8-bit colour
    * using the default 256-colour palette. No transparency.
    * @param src the source image to convert
    * @return a copy of the source image with an 8-bit colour depth
    */
    public static java.awt.image.BufferedImage reduceTo256(java.awt.image.BufferedImage src) 
    {
        java.awt.image.BufferedImage interm = new java.awt.image.BufferedImage(
            src.getWidth(), src.getHeight(),
            java.awt.image.BufferedImage.TYPE_BYTE_INDEXED
            );
        java.awt.image.ColorConvertOp cco = new java.awt.image.ColorConvertOp(
            src.getColorModel().getColorSpace(),
            interm.getColorModel().getColorSpace(),
            null
            );
        cco.filter(src, interm);
        java.awt.image.BufferedImage dest = new java.awt.image.BufferedImage(
                src.getWidth(), src.getHeight(), 
                java.awt.image.BufferedImage.TYPE_INT_ARGB); 
        java.awt.Graphics2D g2 = dest.createGraphics();
        g2.drawImage(interm, 0, 0, null);
        g2.dispose();
  //      copyAlpha(src,dest);
        return dest;
    }
     /**
    * Converts the source image to 8-bit colour
    * using the Aurora colour palette. 
    * 
    * @param src the source image to convert
    * 
    * @return a copy of the source image with an aurora 8-bit colour depth
    */
    public static java.awt.image.BufferedImage reduceToAurora(java.awt.image.BufferedImage src) 
    {
        int xs=src.getWidth();
        int ys=src.getHeight();
        java.awt.image.BufferedImage dest = new java.awt.image.BufferedImage(
                xs,ys, 
                java.awt.image.BufferedImage.TYPE_INT_ARGB); 
        int[] destRgb= dest.getRGB(0,0,xs,ys, null, 0, xs);
        int []srcRgb=src.getRGB(0,0,xs,ys, null, 0, xs);
        if (srcRgb.length!=destRgb.length)
            return null;
        int alpha,val,destval;                              // alpha value, source value, final value
        int rd,bd,gd;                                       // individual R G B components
        for (int i=0;i<srcRgb.length;i++)
        {
            val=srcRgb[i];
            alpha=val&0xff000000;                           // keep aplha value
            rd=val>>>16;
            gd=val>>8;
            rd&=0xe0;                                       // 00000000 rrr00000
            rd<<=1;                                         // 0000000r rr000000
            gd&=0xe0;                                       // 00000000 ggg00000
            gd>>>=2;                                        // 00000000 00ggg000
            bd=val&0xe0;                                    // 00000000 bbb00000
            bd>>>=5;                                        // 00000000 00000bbb
            rd|=gd;                                         // 0000000r rrggg000 
            rd|=bd;                                         // 0000000r rrgggbbb
            rd=(ConvertRGB.RGB2AUR_TAB[rd])&0xff;           // convert this rgb to aurora value
            destval=ConvertRGB.AUR2RGB_TAB[rd];             // convert aurora value to rgb.
            destval|=alpha;                                 // keep alpha
            destRgb[i]=destval;
        }
        dest.setRGB(0,0,xs,ys, destRgb, 0, xs);
        return dest;
    }
      /**
    * Converts the source image to 16-bit colour correspondinf to the SMSQE sprite mode 32
    * using the default 256-colour palette. No transparency.
    * @param src the source image to convert
    * @return a copy of the source image with an 8-bit colour depth
    */
    public static java.awt.image.BufferedImage reduceToMode32(java.awt.image.BufferedImage src) 
    {
        
        int xs=src.getWidth();
        int ys=src.getHeight();
        java.awt.image.BufferedImage dest = new java.awt.image.BufferedImage(
                xs,ys, 
                java.awt.image.BufferedImage.TYPE_INT_ARGB); 
        int[] destRgb= dest.getRGB(0,0,xs,ys, null, 0, xs);
        int []srcRgb=src.getRGB(0,0,xs,ys, null, 0, xs);
        if (srcRgb.length!=destRgb.length)
            return null;
        int val;                              // alpha value, source value, final value
        for (int i=0;i<srcRgb.length;i++)
        {
            val=srcRgb[i]&0xfff8fcf8;
            destRgb[i]=val;
        }
        dest.setRGB(0,0,xs,ys, destRgb, 0, xs);
        return dest;
    }
    
    
    private static void copyAlpha(java.awt.image.BufferedImage src,java.awt.image.BufferedImage dest)
    {
        int xs=src.getWidth();
        int ys=src.getHeight();
        if ((xs!= dest.getWidth())|| (ys!=dest.getHeight()))
            return;
        int[] destRgb= dest.getRGB(0,0,xs,ys, null, 0, xs);
        int []srcRgb=src.getRGB(0,0,xs,ys, null, 0, xs);
        if (srcRgb.length!=destRgb.length)
            return ;
        int alpha,destval;
        for (int i=0;i<srcRgb.length;i++)
        {
            alpha=srcRgb[i]&0xff000000;
            destval=destRgb[i]&0xffffff;
            destval|=alpha;
            destRgb[i]=destval;
        }
        dest.setRGB(0,0,xs,ys, destRgb, 0, xs);
    }
}
