package smsqesprites;



/**
 * Object to make a fading reflection of an image.
 * This is achieved by creating a new image combining the source and the reflection.
 * There may be a gap between the two, you can also determine the height of the reflection and its initial opacity.
 * 
 * @author and copyright (c) Wolfgang Lenerz 2013
 */
public class Reflection 
{
    /**
     * Makes a fading reflection of an image.
     * 
     * @param source the source image
     * @param gap the gap between the bottom of the source image and the top of the reflection.
     * @param opacity how opaque the reflection is at the start, as a percentage between 0.0 (transparent) and 1.0 (opaque).
     * @param fadeHeight how high the reflection should be.
     * 
     * @return a new BufferedImage, the combined source and reflection.
     */
    public static final java.awt.image.BufferedImage reflectedImage (java.awt.image.BufferedImage source, int gap,float opacity,float fadeHeight)
    {       
        int imageWidth = source.getWidth();
        int imageHeight = source.getHeight();
        float h=(float)((float)(imageHeight)*fadeHeight);
        int reflectHeight=(int) (h);                        // height of the reflection
        int newHeight=reflectHeight+imageHeight+gap;        // height of combined image
        
        java.awt.image.BufferedImage finalImage = new java.awt.image.BufferedImage(imageWidth, newHeight, java.awt.image.BufferedImage.TYPE_INT_ARGB);
        java.awt.Graphics2D g2D = finalImage.createGraphics();
        g2D.drawImage(source, null, 0,0);                   // draw source into final image
        
        float op=(float)((float)(255)*opacity);             // max opacity (opacity of first reflected row at bottom of source image)
        if (op==0.0)
            return finalImage;                              // no opacity at all? - no reflection, just leave
        int opVal=(int)op;                                  // max opacity as int
        opVal&=0xff;                                        // opacity = alpha value
        int opValuns=opVal;                                 // unshifted alpha value : 000A
        opVal<<=24;                                         // A000
        float opSteps=(float)(op/h);                        // steps in opacity
        int []rgbFinal=new int[imageWidth * newHeight];
        finalImage.getRGB(0,0,imageWidth,imageHeight, rgbFinal, 0, imageWidth);// get the data from the source image into the array
        int sourceline=((imageHeight-1)*imageWidth);
        for (int lines=imageHeight+gap;lines<newHeight;lines++)// lines is the destination
        {
            int destStart=lines*imageWidth;                 // destination line
            for (int i=0;i<imageWidth;i++)
            {
                int rgb=rgbFinal[sourceline+i];             // line in source
                int alpha=(rgb&0xff000000)>>>24;            // source alpha value
                rgb&=0xffffff;
                if (alpha>opValuns)                         // the reflection alpha value mustn't be bigger than the one in the source
                {
                    rgb|=opVal;
                }
                else
                    rgb|=(alpha<<24); 
                rgbFinal[destStart+i]=rgb;
            }
            sourceline-=imageWidth;                         // new sourceline
            op-=opSteps;
            opVal=(int)op;
            opVal&=0xff;
            opValuns=opVal;
            opVal<<=24;
        }
        finalImage.setRGB(0,0, imageWidth,newHeight, rgbFinal, 0, imageWidth);
        return finalImage;
    }
}
