/*   
        Compressed data starts with the bytes 'RLEx', with 'x' being
        either 1, 2 or 4. This is the item size the algorithm is working with.
        8 bit RLE compression of 32 bit data wouldn't yield in good results,
        therefore the algoritm can also work on 16 bit or 32 bit data.
        After the ID there's one long word containing the size of the data in
        uncompressed form. After that the compressed data itself is following.

        The compressed data always consists of one byte and one or more items.
        If the leading byte is in the range of 0<=x<128 then x+1 uncompressed
        items are following. Otherwise only one item is following, which
        represents 257-x times the same item in the uncompressed data.
*/

package smsqesprites;

/**
 * This compresses an array of ints, shorts or bytes into a Bytebyffer according to the SMSQE sprite specifications.
 * There are three overloaded public methods, for arrays of ints, shorts and bytes.
 * 
 * @author and copyright Wolfgang Lenerz 2013
 */
public class RleEncoder 
{
    /* *************************************** RLE compression for an array of ints ***************************************************/    
    
    /**
     * Do an SMSQE sprite style RLE compression for an array of ints.
     * 
     * @param array the array of ints to compress.
     * 
     * @return a java.nio.ByteBuffer with the (possibly compressed) data. If the data is not usefully compressible, the data isn't compressed.
     * Thus, the result should be checked for the compression flag (RLEx) in the first int. 
     * The size of the buffer will be even, there may thus be an extra (empty) padding byte at the end.
     */
    public static java.nio.ByteBuffer doRLECompression(int[] array)
    {
        int totcount=getCompressedSize(array);              // compressed size of array
        int arrsize=array.length*4;                         // uncompressed size of array
        java.nio.ByteBuffer buf;
        if (arrsize<=totcount)                              // if the compressed size would be bigger than the uncompressed size, use uncompressed array.
        {
            buf=java.nio.ByteBuffer.allocateDirect(arrsize);
            for (int i:array)
            {
                buf.putInt(i);
            }
        }
        else
        {
            if ((totcount&1)!=0)
                totcount++;                                 // make sure buffer size is even
            buf=java.nio.ByteBuffer.allocateDirect(totcount);
            int count;
            buf.putInt(0x524c4534);                         // "RLE4"
            buf.putInt(arrsize);                            // uncompressed size
            int startptr=0;
            while (startptr<array.length-1)
            {
                count=countUnequalInts(startptr,array);     // get number of unequals ints
                if (count!=0)
                {
                    writeUnequalInts(count,startptr,array,buf); // write out number of unequals
                    startptr+=count;
                }
                count=countEqualInts(startptr,array);       // get number of equal ints
                if (count!=0)
                {
                    writeEqualInts(count,startptr,array,buf);   // write out number of unequals
                    startptr+=count;
                }
            }
            buf.put((byte)0);
            buf.putInt(array[array.length-1]);
        }
        return buf;
    }
    
    /**
     * Checks whether compressing an array of ints would make sense.
     * 
     * @param array
     * 
     * @return the number of bytes necessary for a bytebuffer long enough to hold the compressed data.
     * Obviously, if this were larger than a bytebuffer holding the uncompressed data, don't compress...
     */
    private static int getCompressedSize(int[] array)
    {
        int startptr=0;
        int totcount=0;
        int count;
        while (startptr<array.length-1)
        {
            count=countUnequalInts(startptr,array);         // get number of unequals elements
            if (count!=0)
            {
                startptr+=count;                            // where to start counting equal elements
                totcount+=(count*4);                        // running total : each unequal element gets 4 bytes
                totcount+=(count/127)+1;                    // plus number of leading bytes for header : one per 127 elements
            }
            count=countEqualInts(startptr,array);           // get number of equal ints
            if (count!=0)
            {
                startptr+=count;                            // where to restart counting unequal elements
                count=(count/127)+1;                        // nbr of times this is put in
                totcount+=(count*5);                        // 5 bytes each time (leading byte, 4 bytes element)
            }
        }
        return totcount+5+4+4;                              // last element + header + uncompressed size
    }
    
    /**
     * Simply count all unequal ints from the startposition onwards.
     * 
     * @param startptr where to start counting in the array.
     * @param array the array of ints.
     * 
     * @return nbr of unequal ints.
     */
    private static int countUnequalInts(int startptr,int[] array)
    {
        int running=startptr;
        int count=0;
        while (running<array.length-1) 
        {
            if (array[running]==array[running+1])
                break;
            count++;
            running++;
        }
        //return running-startptr;
        return count;
    }
    /**
     * Simply count all equal ints from the startposition onwards.
     * 
     * @param startptr where to start counting in the array.
     * @param array the array of ints.
     * 
     * @return nbr of equal ints.
     */
    private static int countEqualInts(int startptr,int[] array)
    {
        int running=startptr;
        int count=0;
        while (running<array.length-1) 
        {
            if (array[running]!=array[running+1])
                break;
            count++;
            running++;
        }
        return count;
    }
    /**
     * Takes ints from the source array and writes them, uncompressed, into the destination ByteBuffer.
     * The format is : byte (=number of ints to follow (no more than 127)) followed by the ints.
     * 
     * @param count how many ints to write
     * @param startptr where to start taking data from in the source array.
     * @param array the source array.
     * @param buf the destination ByteBuffer. Data is written at the current position.
     */
    private static void writeUnequalInts(int count,int startptr,int[] array,java.nio.ByteBuffer buf)
    {
        try
        {
            int nbr=0;
            int start=startptr;
            while (count>0)
            {
                if (count>127)
                {
                    nbr=127;                                    // don't write more than 127 ints without a size byte.
                }
                else
                    nbr=count;                                  // nbr of ints to treat now
                count-=nbr;                                     // new count
                buf.put((byte)(nbr-1));                         // nbr of ints to follow-1
                for (int i=0;i<nbr;i++)
                {
                    buf.putInt(array[start++]);                 // leave the start++ as is (don't use i !!!!)
                }
            }
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
    }
    /**
     * /**
     * Takes an int from the source array and writes it into the destination ByteBuffer flagging it as compressed data.
     * The format is : byte (=number of times the int to follow is repeated (no more than 127)) followed by the int.
     * 
     * @param count how many ints to write
     * @param startptr where to start taking data from in the source array.
     * @param array the source array.
     * @param buf the destination ByteBuffer. Data is written at the current position.
     */
    private static void writeEqualInts(int count,int startptr,int[] array,java.nio.ByteBuffer buf)
    {
        int nbr=0;
        while (count>0)
        {
            if (count>127)
            {
                nbr=127;                                    // can't havee more than 127 repetitions
            }
            else
                nbr=count;                                  // nbr of ints to treat now
            count-=nbr;                                     // new count
            nbr=257-nbr;                                    // format required fr the numbers byte
            buf.put((byte)nbr);                             // write nbr of ints to follow
            buf.putInt(array[startptr]);                    // write int
        }
    }
    
    
    /* *************************************** RLE compression for an array of shorts ***************************************************/    
    
    
     /**
     * Do an SMSQE sprite style RLE compression for an array of shorts.
     * 
     * @param array the array of bytes to compress.
     * 
     * @return a java.nio.ByteBuffer with the (possibly compressed) data. If the data is not usefully compressible, the data isn't compressed.
     * Thus, the result should be checked for the compression flag (RLEx) in the first int. The size of the buffer will be even, there may thus be
     * an extra (empty) padding byte at the end.
     */
    public static final java.nio.ByteBuffer doRLECompression(short[] array)
    {
        int totcount=getCompressedSize(array);              // compressed size of array in bytes
        int arrsize=array.length*2;                         // uncompressed size of array in bytes
        java.nio.ByteBuffer buf;
        if (arrsize<=totcount)
        {    
            buf=java.nio.ByteBuffer.allocateDirect(arrsize);
            for (short i:array)
            {
                buf.putShort(i);
            }
        }
        else
        {
            if ((totcount&1)!=0)
                totcount++;                                 // make sure buffer size is even
            buf=java.nio.ByteBuffer.allocateDirect(totcount);
            int count;
            buf.putInt(0x524c4532);                         // "RLE2"
            buf.putInt(arrsize);                            // uncompressed size
            int startptr=0;
            while (startptr<array.length-1)
            {
                count=countUnequalShorts(startptr,array);   // get number of unequals ints
                if (count!=0)
                {
                    writeUnequalShorts(count,startptr,array,buf);    // write out number of unequals
                    startptr+=count;
                }
                count=countEqualShorts(startptr,array);     // get number of equal ints
                
                if (count!=0)
                {
                    writeEqualShorts(count,startptr,array,buf);  // write out number of unequals
                    startptr+=count;
                }
            }
            buf.put((byte)0);
            buf.putShort(array[array.length-1]);
        }
        return buf;
    }
    /**
     * Checks whether compressing an array of shorts would make sense.
     * 
     * @param array
     * 
     * @return the number of bytes necessary for a ByteBuffer long enough to hold the compressed data.
     * Obviously, if this were larger than a ByteBuffer holding the uncompressed data, don't compress...
     */
    private static int getCompressedSize(short[] array)
    {
        int startptr=0;
        int totcount=0;
        int count;
        while (startptr<array.length-1)
        {
            count=countUnequalShorts(startptr,array);       // get number of unequals elements
            if (count!=0)
            {
                startptr+=count;                            // where to start counting equal elements
                totcount+=(count*2);                        // running total : each unequal element gets 2 bytes
                totcount+=(count/127)+1;                    // plus number of leading bytes for header : one per 127 elements
            }
            count=countEqualShorts(startptr,array);         // get number of equal shorts
            if (count!=0)
            {
                startptr+=count;                             // where to restart counting unequal elements
                count=(count/127)+1;                         // nbr of times this is put in
                totcount+=(count*3);                         // 3 bytes each time (leading byte, 2 bytes element)
            }
        }
        return totcount+3+4+4;                              // last element + header + uncompressed size
    }
    
    /**
     * Simply count all unequal shorts from the start position onwards.
     */
    private static int countUnequalShorts(int startptr,short[] array)
    {
        int running=startptr;
        int count=0;
        while (running<array.length-1) 
        {
            if (array[running]==array[running+1])
                break;
            count++;
            running++;
        }
        return count;
    }
    /**
     * Simply count all equal shorts from the start position onwards.
     */
    private static int countEqualShorts(int startptr,short[] array)
    {
        int running=startptr;
        int count=0;
        while (running<array.length-1) 
        {
            if (array[running]!=array[running+1])
                break;
            count++;
            running++;
        }
        return count;
    }
    
    private static void writeUnequalShorts(int count,int startptr,short[] array,java.nio.ByteBuffer buf)
    {
        try
        {
        int nbr=0;
        int start=startptr;
        while (count>0)
        {
            if (count>127)
            {
                nbr=127;
            }
            else
                nbr=count;                                  // nbr of ints to treat now
            count-=nbr;                                     // new count
            buf.put((byte)(nbr-1));                         // nbr of ints to follow-1
            for (int i=0;i<nbr;i++)
            {
                buf.putShort(array[start++]);               // leave the start++ as is (don't use i !!!!)
            }
        }
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
    }
    
    private static void writeEqualShorts(int count,int startptr,short[] array,java.nio.ByteBuffer buf)
    {
        int nbr=0;
        while (count>0)
        {
            if (count>127)
            {
                nbr=127;
            }
            else
                nbr=count;                                  // nbr of ints to treat now
            count-=nbr;                                     // new count
            nbr=257-nbr;
            buf.put((byte)nbr);                             // nbr of ints to follow
            buf.putShort(array[startptr]);
        }
    }

 
    
    /* *************************************** RLE compression for an array of bytes ***************************************************/    
    
    
     /**
     * Do an SMSQE sprite style RLE compression for an array of bytes.
     * 
     * @param array the array of bytes to compress.
     * 
     * @return a java.nio.ByteBuffer with the (possibly compressed) data. If the data is not usefully compressible, the data isn't compressed.
     * Thus, the result should be checked for the compression flag (RLEx) in the first int. The size of the buffer will be even, there may thus be
     * an extra (empty) padding byte at the end.
     */
    public static final java.nio.ByteBuffer doRLECompression(byte[] array)
    {
        int totcount=getCompressedSize(array);              // compressed size of array
        int arrsize=array.length;                           // uncompressed size of array
        java.nio.ByteBuffer buf;
        if (arrsize<=totcount)
        {
            if ((arrsize&1)!=0)
                arrsize++;       
            buf=java.nio.ByteBuffer.allocateDirect(arrsize);
            for (byte i:array)
            {
                buf.put(i);
            }
        }
        else
        {
            if ((totcount&1)!=0)
                totcount++;                                 // make sure buffer size is even
            buf=java.nio.ByteBuffer.allocateDirect(totcount);
            int count;
            buf.putInt(0x524c4531);                         // "RLE1"
            buf.putInt(arrsize);                            // uncompressed size
            int startptr=0;
            while (startptr<array.length-1)
            {
                count=countUnequalBytes(startptr,array);    // get number of unequals ints
                if (count!=0)
                {
                    writeUnequalBytes(count,startptr,array,buf);// write out number of unequals
                    startptr+=count;
                }
                count=countEqualBytes(startptr,array);      // get number of equal ints
                if (count!=0)
                {
                    writeEqualBytes(count,startptr,array,buf);// write out number of unequals
                    startptr+=count;
                }
            }
            buf.put((byte)0);
            buf.put(array[array.length-1]);
        }
        return buf;
    }
    /**
     * Checks whether compressing an array of bytes would make sense.
     * 
     * @param array
     * 
     * @return the number of bytes necessary for a ByteBuffer long enough to hold the compressed data.
     * Obviously, if this were larger than a ByteBuffer holding the uncompressed data, don't compress...
     */
    private static int getCompressedSize(byte[] array)
    {
        int startptr=0;
        int totcount=0;
        int count;
        while (startptr<array.length-1)
        {
            count=countUnequalBytes(startptr,array);        // get number of unequals elements
            if (count!=0)
            {
                startptr+=count;                                // where to start counting equal elements
                totcount+=(count);                              // running total : each unequal element gets 1 byte
                totcount+=(count/127)+1;                        // plus number of leading bytes for header : one per 127 elements
            }
            count=countEqualBytes(startptr,array);          // get number of equal ints
            if (count!=0)
            {
                startptr+=count;                                // where to restart counting unequal elements
                count=(count/127)+1;                            // nbr of times this is put in
                totcount+=(count*2);                            // 2 bytes each time (leading byte, 1 byte element)
            }
        }
        return totcount+2+4+4;                              // last element + header + uncompressed size
    }
    
    /**
     * Simply count all unequal bytes from the start position onwards.
     */
    private static int countUnequalBytes(int startptr,byte[] array)
    {
        int running=startptr;
        int count=0;
        while (running<array.length-1) 
        {
            if (array[running]==array[running+1])
                break;
            count++;
            running++;
        }
        //return running-startptr;
        return count;
    }
    /**
     * Simply count all equal bytes from the start position onwards.
     */
    private static int countEqualBytes(int startptr,byte[] array)
    {
        int running=startptr;
        int count=0;
        while (running<array.length-1) 
        {
            if (array[running]!=array[running+1])
                break;
            count++;
            running++;
        }
        return count;
    }
    
    private static void writeUnequalBytes(int count,int startptr,byte[] array,java.nio.ByteBuffer buf)
    {
        try
        {
        int nbr=0;
        int start=startptr;
        while (count>0)
        {
            if (count>127)
            {
                nbr=127;
            }
            else
                nbr=count;                                  // nbr of ints to treat now
            count-=nbr;                                     // new count
            buf.put((byte)(nbr-1));                         // nbr of ints to follow-1
            for (int i=0;i<nbr;i++)
            {
                buf.put(array[start++]);                 // leave the start++ as is (don't use i !!!!)
            }
        }
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
    }
    
    private static void writeEqualBytes(int count,int startptr,byte[] array,java.nio.ByteBuffer buf)
    {
        int nbr=0;
        while (count>0)
        {
            if (count>127)
            {
                nbr=127;
            }
            else
                nbr=count;                                  // nbr of ints to treat now
            count-=nbr;                                     // new count
            nbr=257-nbr;
            buf.put((byte)nbr);                             // nbr of ints to follow
            buf.put(array[startptr]);
        }
    }
}
