/*
 
SMSQE sprite definition info.
* 
The sprite definition has been extensively modified, but in such a way
that it should continue to be compatible with older sprites.

The sprite header is now as follows:

pto_form $00  byte  sprite mode
         $01  byte  colour mode / system sprite number
pto_vers $02  byte  dynamic sprite version number
pto_ctrl $03  byte  sprite control
pto_xsiz $04  word  X size
pto_ysiz $06  word  Y size
pto_xorg $08  word  X offset
pto_yorg $0a  word  Y offset
pto_cpat $0c  long  relative pointer to colour pattern
pto_mask $10  long  relative pointer to mask/alpha channel
pto_nobj $14  long  relative pointer to next object

pto_opts $18  long  OPTIONAL (relative pointer to) options
pto_blk  $1c  long  OPTIONAL relative pointer to sprite block


(the keys can be found in DEV8_keys_qdos_pt)



B - Sprite control byte:
------------------------
The sprite control bit is formatted as follows

        %mpao0xcc

        cc
        stands for a cache version number. Programs can increment this
        value to signal the cache that the sprite has changed. A special
        value is pto.fupd (force update, %11), which causes the system to
        never use the cached version.
        key: pto.cver

        a = 32
        flags whether the sprite uses an alpha channel instead of a mask
        (see below)
        key: pto.alph

        m and p 64/128
        signal whether the pattern (p) or the mask (m) is compressed (see below)
        key: pto.pcmp and pto.mcmp

        x
        signals that the optional pointer to a sprite block is present (bit x=1)
        or is not present (bit x=0) (see below)
        key pto..blk

        o signals that options are present (=1) or not (=0)(see below)
        key pto..opt
        currently unused

D - RLE compression:
--------------------
        Both pattern and mask/alpha channel can be compressed using a simple
        RLE (run length encoding) algorithm. This is usefull with data that
        is largely homogene, which is often the case with masks.
        Compressed data must be signaled in the sprite control byte
        (pto.pcmp, pto.mcmp) and starts with the bytes 'RLEx', with 'x' being
        either 1, 2 or 4. This is the item size the algorithm is working with.
        8 bit RLE compression of 32 bit data wouldn't yield in good results,
        therefore the algoritm can also work on 16 bit or 32 bit data.
        After the ID there's one long word containing the size of the data in
        uncompressed form. After that the compressed data itself is following.

        The compressed data always consists of one byte and one or more items.
        If the leading byte is in the range of 0<=x<128 then x+1 uncompressed
        items are following. Otherwise only one item is following, which
        represents 257-x times the same item in the uncompressed data.
*/
package smsqesprites;


/**
 * Conversion of image to an SMSQ/E sprite (modes 4,16,32,64 supported).
 * 
 * @author and copyright (c) Wolfgang Lenerz 2013
 */
public class SMSQESprite 
{
    /**
     * This makes the sprite in SMSQ/E format.
     * @param f the file to save to
     * @param bimage the iamge to convert into a sprite
     * @param mode the mode of the sprite : 4,16,32,64
     */
    public static void makeSprite(java.io.File f,java.awt.image.BufferedImage bimage,int mode,boolean noCache,boolean fullMask)
    {
        if (bimage==null)
            return;
        try
        {
            int imageWidth = bimage.getWidth();
            int imageHeight = bimage.getHeight();
            int dataSize=imageWidth*imageHeight;
            if (mode==4)
                bimage=ColourReduction.reduceToMode4(bimage);
            int []rgbArray=bimage.getRGB(0,0,imageWidth,imageHeight, null, 0, imageWidth);// get the data from the image into the array with ARGB colour values
            byte[]alphaArray=new byte[dataSize];            // the alpha values
            byte alpha;
            int col;
            for (int i=0;i<dataSize;i++)                    // make the array with the alpha mask values
            {
                if (fullMask)
                    alphaArray[i]=-1;
                else
                {
                    col=rgbArray[i];                        // ARGB
                    alpha=(byte)(col>>>24);                 // 000A
                    alphaArray[i]=alpha;
                }                            
            }
            java.nio.ByteBuffer dataBuffer=null;            // will contain converted (possibly compressed) SMSQE style colour data
            java.nio.ByteBuffer alphaBuffer=null;           // will contain converted (possibly compressed) SMSQE style mask data
            switch (mode)
            {
                case 4:
                    dataBuffer=ConvertRGB.makeMode4Pattern(rgbArray,imageWidth,imageHeight);
                    alphaBuffer=ConvertRGB.makeMode4Mask(rgbArray,imageWidth,imageHeight,0);
                    break;
                case 16:
                    byte[] mode16=ConvertRGB.convertToMode16(rgbArray,imageWidth,imageHeight);
                    dataBuffer=RleEncoder.doRLECompression(mode16);// now we have the data compressed (or not) in SMSQE format
                    break;
                case 32:
                    short[] mode32=ConvertRGB.convertToMode32(rgbArray,imageWidth,imageHeight);
                    dataBuffer=RleEncoder.doRLECompression(mode32);// now we have the data compressed (or not) in SMSQE format
                    break;
                case 64 :
                    for (int i=0;i<dataSize;i++)
                    {
                        col=rgbArray[i];                    // ARGB
                        col<<=8;                            // RGB0
                        rgbArray[i]=col;                    // SMSQE mode 64 expects data as RGB0
                    }
                    dataBuffer=RleEncoder.doRLECompression(rgbArray);// now we have the data compressed (or not) in SMSQE format
                    break;
            }
            int flag=0;
            if (dataBuffer==null)
                return;                                     // what?????
            if ((dataBuffer.getInt(0)&0xffffff00)==0x524c4500)// is data RLE compressed?
            {
                flag=64;                                    // yes, flag for compressed pattern
            }
            if (alphaBuffer == null)
            {
                alphaBuffer=RleEncoder.doRLECompression(alphaArray);// get data for alpha mask, may be compressed
                flag+=32;                                    // show that we have an alpha mask
            }
            if (alphaBuffer != null && (alphaBuffer.getInt(0)&0xffffff00)==0x524c4500)// is mask RLE compressed?
            {
                flag+=128;                                  // flag for compressed mask
            }
            java.nio.ByteBuffer header=java.nio.ByteBuffer.allocateDirect(24); // space for sprite header  
            if (mode==4)
                header.putShort((short)0x100);              // mode 4 sprite header
            else
            {
                header.put((byte)2);                        // flag for GD2 sprite + RGB sprite
                header.put ((byte)mode);                    // flag for mode of sprite 
                if (noCache)
                    flag+=3;
                header.put(3,(byte)(flag));                 // with alpha (& compressed pattern & mask, possibly)
            }
           
            header.putShort(4,(short)(imageWidth&0xffff));  // uncompressed image width
            header.putShort(6,(short)(imageHeight&0xffff)); // uncompressed height (8 to 11 remain 0 (origins))
            header.put(15,(byte)12);                        // relative pointer to colour data (right after the header)
            col=dataBuffer.capacity()+8;                    // the pattern data is this long
            header.putInt(16,col);                          // relative pointer to mask (rel ptr to next object remains 0)
            saveSprite(header,dataBuffer,alphaBuffer,f);    // write it out to the file
        } 
        catch (Exception e) 
        {
            e.printStackTrace();
        }
    }
  
     /**
     * Saves the sprite. to a file
     * @param header the buffer with the header.
     * @param bufrgb the buffer with the RGB data.
     * @param alpha  The buffer with the mask data.
     * @param fe the file to save to.
     */
    private static void saveSprite(java.nio.ByteBuffer header,java.nio.ByteBuffer bufrgb ,java.nio.ByteBuffer alpha,java.io.File fe)
    {
        try (java.nio.channels.FileChannel file = new java.io.FileOutputStream(fe).getChannel())
        {
            header.position(0);
            file.write(header);
            bufrgb.position(0);
            file.write(bufrgb);
            alpha.position(0);
            file.write(alpha);
        } 
        catch (Exception e) 
        {
            javax.swing.JOptionPane.showMessageDialog(null,"Error writing the sprite!", e.getMessage(), javax.swing.JOptionPane.ERROR_MESSAGE);
        }  
    }
}
