package smsqesprites;
/**
 * The panel with the image. Based on a JPanel.
 * It may create a reflection by making a second image (combining the source image and the reflection).
 * 
 * @author and copyright (c) Wolfgang Lenerz 2013
 */
public class SpritePanel extends javax.swing.JPanel
{
    private java.awt.image.BufferedImage image=null;        // the basic (source) image
    private java.awt.image.BufferedImage reflectedImage=null;// the image with reflection
    private java.awt.image.BufferedImage previewImage=null;// the image once it's changed into the QL modes
    private float opacity = 0.5f;
    private float fadeHeight = 0.5f;                        // initial values for opacity and height of reflection
    private int gap=2;
    private boolean doReflection=false;                     // draw reflection ?
    private int minX=504;
    private int minY=244;                                   // mminimum x and y sizes
    private java.awt.Dimension minDim=new java.awt.Dimension(minX,minY);
    private boolean doPreview;
    
    /**
     * Creates the SpritePanel.
     */
    public SpritePanel()
    {
        super();
       // this.image=testImage();
        this.setPreferredSize(new java.awt.Dimension(400,400));
    }
    
    /**
     * Creates the SpritePanel with an initial image.
     * 
     * @param filename the name of the image file to load.
     */
     public SpritePanel(String filename)
    {
        this();
        loadImage(filename);
    }
     
    /**
     * Paints the component.
     * 
     * Paints either the source image or the reflection image.
     * 
     * @param g the graphics object to paint on.
     */
    @Override
    public void paintComponent(java.awt.Graphics g) 
    {
        super.paintComponent(g);
        int width=this.getWidth();
        int height=this.getHeight();
        
        int x=0;
        int y=0;
        if (this.reflectedImage!=null && this.doReflection)
        {  
            x=this.doPreview?(width-(this.image.getWidth()*2+10))/2:(width-this.image.getWidth())/2;
            y=(height-this.reflectedImage.getHeight())/2;
            g.drawImage(this.reflectedImage, x,y,null);
        }
        else if (this.image!=null)
        {
            x=this.doPreview?(width-(this.image.getWidth()*2+10))/2:(width-this.image.getWidth())/2;;
            y=(height-this.image.getHeight())/2;
            g.drawImage(this.image, x,y,null);
        }
        if (this.doPreview && this.previewImage!=null && (this.image!=null))
        {
            x+=this.image.getWidth()+10;
            g.drawImage(this.previewImage, x,y,null);
        }
    }
    
    /**
     * Gets the minimum size of the panel.
     * 
     * @return the size
     */
    @Override
    public java.awt.Dimension getMinimumSize()
    {
        return this.minDim;
    } 
    /**
     * Gets the preferred size of the panel.
     * 
     * @return the size
     */
    @Override
    public java.awt.Dimension getPreferredSize()
    {
       return this.minDim;
    }
    
    
    /**
     * Loads the image.
     * 
     * @param filename the path and name of the file to load.
     */
    private boolean loadImage(String filename)
    {
        if (filename!=null && !filename.isEmpty())
            return loadImage (new java.io.File (filename));
        else
            return false;
    }  
    /**
     * Loads the image.
     * 
     * @param f the file to load.
     */
    private boolean loadImage (java.io.File f)
    {
        if (f!=null)
        {
            try 
            {
                this.image = javax.imageio.ImageIO.read(f);
                if (this.image!=null)
                {
                    changeSize(this.image.getWidth(),this.image.getHeight());
                    return true;
                }
            }
            catch ( Exception e ) 
            {
              return false;
            }
        }
        return false;
    }
    /**
     * Checks whether the size of the panel must be changed.
     * Only needs to be changed if image doesn't fit.
     * 
     * @param newX the new x size of the image to be displayed.
     * @param newY the new y size of the image to be displayed.
     */
    private void changeSize(int newX,int newY)
    {
        boolean changeSize=false;
        int xs=this.getWidth();
        if (newX<this.minX)
            newX=this.minX;
        if (newX<xs)
            newX=xs;
        if (newX>xs)
            changeSize=true;
        int ys=this.getHeight();
        if (newY<this.minY)
            newY=this.minY;
        if (newY<ys)
            newY=ys;
        if (newY>ys)
            changeSize=true;
        if (changeSize)
        {
            this.minDim.setSize(newX, newY);
            this.setSize(this.minDim);
            this.invalidate();
        }
    }

    /**
     * Adds a reflection to the source image.
     */
    public void addReflection()
    {
        if (this.doReflection)
        {
            this.reflectedImage=Reflection.reflectedImage(this.image, gap, this.opacity,this.fadeHeight);
            changeSize(this.reflectedImage.getWidth(),this.reflectedImage.getHeight());
        }
        this.repaint();
    }
    
    /**
     * Sets whether a reflection is to be shown or not.
     * 
     * @param doReflection 
     */
    public void showReflection(boolean doReflection)
    {
        this.doReflection=doReflection;
        addReflection();
    }
    
    /**
     * Loads and shows the source image.
     * If successful,there will be no reflection, nor preview.
     * 
     * @param f the file with the source image to load..
     */
    public boolean setImage (java.io.File f)
    {
        if (loadImage (f))
        {
            this.repaint();
            this.doPreview=false;
            this.doReflection=false;
            this.reflectedImage=null;
            this.previewImage=null;
            return true;
        }
        return false;
    }
    
    /**
     * Sets the value for the opacity.
     * @param opa a float between between 0.0 (transparent) and 1.0 (max opacity)
     */
    public void setOpacity(float opa)
    {
        if (opa>1.0)
            opa=(float)1.0;
        if (opa<0.0)
            opa=(float)0.0;
        this.opacity=opa;
        addReflection();
    }
    /**
     * Sets the height of the reflection.
     * 
     * @param opa a float between between 0.0 (no reflection) and 1.0 (big reflection).
     */
    public void setFadeHeight(float opa)
    { if (opa>1.0)
            opa=(float)1.0;
        if (opa<0.0)
            opa=(float)0.0;
        this.fadeHeight=opa;
        addReflection();
    }
    
    /**
     * Sets the gap between the source image and its reflection.
     * 
     * @param gapp a positive integer - the gap between the image and its refelction, in pixels.
     */
    public void setGap(int gapp)
    {
        if (gapp<0)
            gapp=0;
        this.gap=gapp;
        addReflection();
    }
    
    /**
     * Makes an SMSQE sprite out of the image.
     * 
     * @param f the file to save the sprite to.
     * @param mode which mode the sprite is to have : 16, 32 or 64 (corresponding SMSQE sprite modes).
     */
    public void makeSprite(java.io.File f,int mode,boolean noCache,boolean fullMask)
    {
        if (this.doReflection)
            SMSQESprite.makeSprite(f,this.reflectedImage,mode,noCache,fullMask);
        else
            SMSQESprite.makeSprite(f,this.image,mode,noCache,fullMask);
    }
   
    /**
     * Redices the colours to what would be shown in the corresponding SMSQE sprite mode.
     * @param colours the colour mode wished:
     * 
        0       1 bit black&white
        3       1 bit palette mapped
        4       2 bit fixed gr palette
        7       2 bit palette mapped
        8       4 bit fixed irgb palette
        15      4 bit palette mapped
        16      8 bit fixed palette (equals Aurora palette)
        31      8 bit palette mapped
        32      16 bit QPC/QXL %gggbbbbbrrrrrggg format
        33      16 bit Q40 %gggggrrrrrbbbbbw format
        64      32 bit $RRGGBB00 format
     */
    private void reduceColours(int colours)
    {
        this.previewImage=null;
        java.awt.image.BufferedImage myim=this.doReflection?this.reflectedImage:this.image;
        switch(colours)
        {
            case 4:
                this.previewImage=ColourReduction.reduceToMode4(myim);
                break; 
            case 16:
                this.previewImage=ColourReduction.reduceToAurora(myim);
                break;
            case 32:
                this.previewImage=ColourReduction.reduceToMode32(myim);
                break;
        }
        this.repaint();
    }
    
    public boolean setPreview(boolean doPrev,int previewMode)
    {
        if (this.image==null)
            return false;
        this.doPreview=doPrev;
        if (doPrev)
            reduceColours(previewMode);
        else
            this.repaint();
        return doPrev;
    }
    
    private java.awt.image.BufferedImage testImage()
    {
        int xs=29;
        int ys=16;
        java.awt.image.BufferedImage dest = new java.awt.image.BufferedImage(
                29, 16, // match source
                java.awt.image.BufferedImage.TYPE_INT_ARGB); 
        int[] destRgb= dest.getRGB(0,0,xs,ys, null, 0, xs);
        for (int i=0;i<destRgb.length;i++)
            destRgb[i]=0xffff0000;
        for (int i=0;i<ys;i++)
        {
            
            for (int j=0;j<5;j++)
                destRgb[i*xs+j]=0xff2020f8 ;  
            for (int j=5;j<8;j++)
                destRgb[i*xs+j]=0xff000000 ;  
            
            for (int j=8;j<13;j++)
                destRgb[i*xs+j]=0xff2020f8 ;  
            for (int j=13;j<16;j++)
                destRgb[i*xs+j]=0xff000000 ; 
            
             for (int j=16;j<21;j++)
                destRgb[i*xs+j]=0xff2020f8 ;  
            for (int j=21;j<24;j++)
                destRgb[i*xs+j]=0xff000000 ; 
             for (int j=24;j<29;j++)
                destRgb[i*xs+j]=0xff2020f8 ; 
        }
        dest.setRGB(0,0,xs,ys, destRgb, 0, xs);
        return dest;
    }
}
